""" ddeint/__init__.py """

# Credits to Zulko for this module
# https://github.com/Zulko/ddeint
# Some tweaks habe been made for compatibility reasons
__all__ = ["ddeint"]

from .ddeint import ddeint

from .version import __version__
